import re
from enum import Enum
from requests.exceptions import HTTPError
from requests_ratelimiter import LimiterSession

import bibtex_dblp.dblp_data


class InvalidDblpIdException(Exception):
    pass


class BibFormat(Enum):
    """
    Format of DBLP bibtex.
    """

    condensed = "condensed"
    standard = "standard"
    crossref = "crossref"
    condensed_doi = "condensed_doi"

    def bib_url(self):
        """
        Get identifier of format for DBLP urls.
        :return:
        """
        if self is BibFormat.condensed:
            return "0"
        elif self is BibFormat.standard:
            return "1"
        elif self is BibFormat.crossref:
            return "2"
        elif self is BibFormat.condensed_doi:
            return "0"
        else:
            assert False

    def __str__(self):
        return self.value


class DblpSession:
    """
    Keeps track of session for DBLP.
    Needed for rate limiting.
    """

    def __init__(self, wait_time, dblp_base_url="https://dblp.org"):
        """
        Create a session for DBLP.
        :param wait_time: Time in seconds to sleep before retrying.
        :param dblp_base_url: Base URL for DBLP.
        """
        self.base_url = dblp_base_url
        self.wait_time = wait_time

        self.publication_search_url = self.base_url + "/search/publ/api"
        self.publication_bibtex = self.base_url + "/rec/{key}.bib?param={bib_format}"

        self.session = LimiterSession(per_second=1.0 / wait_time, per_minute=60.0 / wait_time, burst=3)

    def perform_request(self, url, params=None, **kwargs):
        """
        Perform a GET request to DBLP.
        :param url: URL to access.
        :param params: Optional parameters.
        :param kwargs: Optional arguments.
        :return: Response.
        :raises: HTTPError if request was unsuccessful.
        """
        response = self.session.get(url, params=params, **kwargs)
        response.raise_for_status()
        return response


def extract_dblp_id(entry):
    """
    Extract DBLP id by either using the biburl if given or trying to use the entry name.
    :param entry: Bibliography entry.
    :return: DBLP id or None if no could be extracted.
    """
    if "biburl" in entry.fields:
        # Try to get id from biburl
        match = re.search(r"http(s?)://dblp.org/rec/(.*).bib", entry.fields["biburl"])
        if match:
            return match.group(2)

    # Try to get id from entry name
    key = entry.key
    if key.startswith("DBLP:"):
        return key[5:]
    return None


def get_bibtex(session, dblp_id, bib_format=BibFormat.condensed):
    """
    Get bibtex entry in specified format.
    :param session: DBLP session.
    :param dblp_id: DBLP id for entry.
    :param bib_format: Format of bibtex export (see BibFormat).
    :return: Bibtex as binary string.
    """
    try:
        resp = session.perform_request(session.publication_bibtex.format(key=dblp_id, bib_format=bib_format.bib_url()))
    except HTTPError as err:
        if err.response.status_code == 404:
            raise InvalidDblpIdException("Invalid DBLP id '{}'".format(dblp_id))
        else:
            raise err

    bibtex = resp.content.decode("utf-8")

    if bib_format == BibFormat.condensed_doi:
        # Also get DOI and insert it into bibtex
        resp = session.perform_request(session.publication_bibtex.format(key=dblp_id, bib_format=BibFormat.standard.bib_url()))
        lines = resp.content.decode("utf-8").split("\n")
        keep_lines = [line for line in lines if line.startswith("  doi")]
        assert len(keep_lines) <= 1
        if keep_lines:
            doi = keep_lines[0][:-1]  # Remove comma
            # Insert into bibtex
            bibtex = bibtex[:-4] + ",\n" + doi + bibtex[-4:]

    if bib_format == BibFormat.condensed or bib_format == BibFormat.condensed_doi:
        # Also insert biburl into bibtex
        assert "biburl" not in bibtex
        biburl = "  biburl = {{https://dblp.org/rec/{}.bib}}".format(dblp_id)
        bibtex = bibtex[:-4] + ",\n" + biburl + bibtex[-4:]

    return bibtex


def search_publication(session, pub_query, max_search_results):
    """
    Search for publication according to given query.
    :param session: DBLP session.
    :param pub_query: Query for publication.
    :param max_search_results: Maximal number of search results to return.
    :return: Search results.
    """
    parameters = dict(q=pub_query, format="json", h=max_search_results)

    resp = session.perform_request(session.publication_search_url, params=parameters)
    results = bibtex_dblp.dblp_data.DblpSearchResults(resp.json())
    assert results.status_code == 200
    return results



import logging


def get_user_input(message):
    """
    Get input from user.
    :param message: Message to display.
    :return: User input.
    """
    return input(message)


def get_user_number(message, val_min=None, val_max=None):
    """
    Get number from user input. If the input is not a valid number, the user is asked again.
    :param message: Message to print.
    :param val_min: Minimal value required for input. If None, no minimal bound is given.
    :param val_max: Maximal value required for output. If None, no maximal bound is given.
    :return: Number given by user.
    """
    while True:
        inp = get_user_input(message)
        try:
            inp_number = int(inp)
        except ValueError:
            logging.error("The input is not a number.")
        else:
            if val_min is not None and inp_number < val_min:
                logging.error("The input must be at least {}.".format(val_min))
            elif val_max is not None and inp_number > val_max:
                logging.error("The input must be at most {}.".format(val_max))
            else:
                # Number is valid
                return inp_number



import os
import pytest

from bibtex_dblp.dblp_api import DblpSession


@pytest.fixture(scope="session")
def dblp_session():
    return DblpSession(wait_time=7)


def bib_path(*paths):
    return os.path.join(os.path.dirname(__file__), "files", *paths)



class DblpSearchResults:
    """
    Results of one search in DBLP.
    """

    def __init__(self, json):
        result = json["result"]
        self.query = result["query"]
        status = result["status"]
        self.status_code = int(status["@code"])
        self.status_text = status["text"]
        hits = result["hits"]
        self.total_matches = int(hits["@total"])
        self.results = []
        if self.total_matches > 0:
            self.results = [DblpSearchResult(hit_json) for hit_json in hits["hit"]]


class DblpSearchResult:
    """
    One possible matched result of a search query in DBLP.
    """

    def __init__(self, json):
        self.score = json["@score"]
        self.publication = DblpPublication(json["info"])


class DblpPublication:
    """
    Publication in DBLP.
    """

    def __init__(self, json):
        self.title = json["title"]
        self.booktitle = json.get("booktitle")
        self.volume = json.get("volume")
        self.venue = json.get("venue")
        self.pages = json.get("pages")
        self.year = int(json.get("year"))
        self.type = json.get("type")
        self.key = json.get("key")
        self.doi = json.get("doi")
        self.ee = json.get("ee")
        self.url = json.get("url")
        self.authors = []
        authors = json.get("authors")
        if authors:
            authors = authors["author"]
            if isinstance(authors, list):
                self.authors = [DblpAuthor(author) for author in authors]
            else:
                self.authors = [DblpAuthor(authors)]

        # Possible additional fields:
        # sub_type, mdate, authors, editors, month, journal, number, chapter, isbn, crossref, publisher, school, citations, series

    def cite_key(self):
        return "DBLP:" + self.key

    def __str__(self):
        s = ", ".join([str(author) for author in self.authors])
        s += ":\n\t"
        s += self.title[:-1] if self.title.endswith(".") else self.title
        s += ", {}".format(self.venue) if self.venue else ""
        s += " ({})".format(self.volume) if self.volume else ""
        s += ", {}".format(self.year)
        if self.ee:
            s += "\n\t{}".format(self.ee)
        return s


class DblpAuthor:
    """
    Author in DBLP.
    """

    def __init__(self, json):
        self.name = json.get("text")
        self.pid = json.get("pid")

    def __str__(self):
        return self.name



#!/usr/bin/env python
"""
Convert DBLP entries in a given bibliography according to the specific DBLP format (condensed, standard or crossref).
"""

import argparse
import logging
from pathlib import Path

import bibtex_dblp.database
from bibtex_dblp.dblp_api import BibFormat, DblpSession


def main():
    parser = argparse.ArgumentParser(description="Convert DBLP entries to specific format (condensed, standard, crossref).")

    parser.add_argument("infile", help="Input bibtex file", type=Path)
    parser.add_argument("--out", "-o", help="Output bibtex file. If no output file is given, the input file will be overwritten.", type=Path, default=None)
    parser.add_argument("--format", "-f", help="DBLP format type to convert into", type=BibFormat, choices=list(BibFormat), default=BibFormat.condensed)
    parser.add_argument("--sleep-time", "-t", help="Sleep time (in seconds) between requests. Can prevent errors with too many requests)", type=int, default=5)

    parser.add_argument("--verbose", "-v", help="print more output", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG if args.verbose else logging.INFO)

    outfile = args.infile if args.out is None else args.out

    bib = bibtex_dblp.database.load_from_file(args.infile)
    session = DblpSession(wait_time=args.sleep_time)
    bib, no_changes = bibtex_dblp.database.convert_dblp_entries(session, bib, bib_format=args.format)
    logging.info("Updated {} entries (out of {}) from DBLP".format(no_changes, len(bib.entries)))
    bibtex_dblp.database.write_to_file(bib, outfile)
    logging.info("Written to {}".format(outfile))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Apply modifications on the bibtex entries.
"""

import argparse
import logging
from pathlib import Path
import re

import bibtex_dblp.database


def main():
    parser = argparse.ArgumentParser(description="Apply modifications on the bibtex file.")

    parser.add_argument("infile", help="Input bibtex file", type=Path)
    parser.add_argument("--out", "-o", help="Output bibtex file. If no output file is given, the input file will be overwritten.", type=Path, default=None)
    parser.add_argument("--no-escape", help="Disable escaping of underscores in URLs", action="store_true")
    parser.add_argument("--no-timestamp", help="Remove timestamp field (if present)", action="store_true")
    parser.add_argument("--no-biburl", help="Remove biburl field (if present)", action="store_true")
    parser.add_argument("--no-bibsource", help="Remove bibsource field (if present)", action="store_true")

    parser.add_argument("--verbose", "-v", help="print more output", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG if args.verbose else logging.INFO)

    outfile = args.infile if args.out is None else args.out

    bib = bibtex_dblp.database.load_from_file(args.infile)
    # Apply modifications
    bib = bibtex_dblp.database.modify_entries(
        bib, remove_escapes=args.no_escape, remove_timestamp=args.no_timestamp, remove_biburl=args.no_biburl, remove_bibsource=args.no_bibsource
    )
    bibtex_dblp.database.write_to_file(bib, outfile)
    if args.no_escape:
        # Need to manually remove escape characters which are automatically added again by pybtex to ensure LaTeX compatibility
        out_lines = []
        for line in outfile.read_text(encoding="utf-8").splitlines(keepends=True):
            # Detect start of a doi or url field
            if re.match(r"\s*(doi|url)\s*=", line, flags=re.IGNORECASE):
                # Remove backslash
                line = re.sub(r"\\_", r"_", line)
            out_lines.append(line)
        outfile.write_text("".join(out_lines), encoding="utf-8")

    logging.info("Written to {}".format(outfile))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Import entry from DBLP according to given search input.
"""

import argparse
import logging
import pyperclip
from pathlib import Path

import bibtex_dblp.database
import bibtex_dblp.dblp_api
import bibtex_dblp.io
from bibtex_dblp.dblp_api import BibFormat, DblpSession


def main():
    parser = argparse.ArgumentParser(description="Import entry from DBLP according to given search input from cli.")

    parser.add_argument(
        "--query", "-q", help="The query to search for the publication. If none is given the query is obtained from CLI input.", type=str, default=None
    )
    parser.add_argument(
        "--bib",
        "-b",
        help="Bibtex file where the imported entry will be appended. If no bibtex file is given, the bibtex is printed to the CLI.",
        type=Path,
        default=None,
    )
    parser.add_argument("--format", "-f", help="DBLP format type to convert into.", type=BibFormat, choices=list(BibFormat), default=BibFormat.condensed)
    parser.add_argument("--max-results", help="Maximal number of search results to display.", type=int, default=30)
    parser.add_argument("--sleep-time", "-t", help="Sleep time (in seconds) between requests. Can prevent errors with too many requests)", type=int, default=5)

    parser.add_argument("--verbose", "-v", help="print more output", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG if args.verbose else logging.INFO)
    max_search_results = args.max_results

    bib = None
    if args.bib is not None:
        # Load bibliography
        bib = bibtex_dblp.database.load_from_file(args.bib)

    if args.query:
        search_words = args.query
    else:
        search_words = bibtex_dblp.io.get_user_input("Give the publication title to search for: ")

    if not search_words:
        print("No search terms. Cancelled.")
        exit(1)

    # Search for publications
    if bib is not None:
        # Check if publication already exists
        bib_result = bibtex_dblp.database.search(bib, search_words)
        if bib_result:
            print("The bibliography already contains the following matches:")
            for i in range(len(bib_result)):
                print("({})\t{}".format(i + 1, bibtex_dblp.database.print_entry(bib_result[i][0])))
            select = bibtex_dblp.io.get_user_number("Select the intended publication (0 to search online): ", 0, len(bib_result))
            if select > 0:
                selected_entry = bib_result[select - 1][0]
                pyperclip.copy(selected_entry.key)
                logging.info("Copied cite key '{}' to clipboard.".format(selected_entry.key))
                exit(0)

    session = DblpSession(wait_time=args.sleep_time)
    search_results = bibtex_dblp.dblp_api.search_publication(session, search_words, max_search_results=max_search_results)
    if search_results.total_matches == 0:
        print("The search returned no matches.")
        exit(1)

    print("The search returned {} matches:".format(search_results.total_matches))
    if search_results.total_matches > max_search_results:
        print("Displaying only the first {} matches.".format(max_search_results))
    for i in range(len(search_results.results)):
        result = search_results.results[i]
        print("({})\t{}".format(i + 1, result.publication))

    # Let user select correct publication
    select = bibtex_dblp.io.get_user_number("Select the intended publication (0 to abort): ", 0, search_results.total_matches)
    if select == 0:
        print("Cancelled.")
        exit(1)

    publication = search_results.results[select - 1].publication
    pub_bibtex = bibtex_dblp.dblp_api.get_bibtex(session, publication.key, bib_format=args.format)
    if args.bib:
        with open(args.bib, "a") as f:
            f.write(pub_bibtex)
        logging.info("Bibtex file appended to {}.".format(args.bib))
        pyperclip.copy(publication.cite_key())
        logging.info("Copied cite key '{}' to clipboard.".format(publication.cite_key()))
    else:
        pyperclip.copy(pub_bibtex)
        logging.info("Selected bibtex entry:\n{}".format(pub_bibtex))
        logging.info("Copied selected bibtex entry to clipboard.")


if __name__ == "__main__":
    main()



from conftest import bib_path

import bibtex_dblp.database
import bibtex_dblp.dblp_api
from bibtex_dblp.dblp_api import BibFormat


def test_convert(dblp_session):
    bib = bibtex_dblp.database.load_from_file(bib_path("ley.bib"))
    bib, no_changes = bibtex_dblp.database.convert_dblp_entries(dblp_session, bib, bib_format=BibFormat.condensed_doi)
    assert no_changes == 9


def test_convert_invalid_id(dblp_session):
    bib = bibtex_dblp.database.load_from_file(bib_path("invalid_id.bib"))
    bib, no_changes = bibtex_dblp.database.convert_dblp_entries(dblp_session, bib, bib_format=BibFormat.condensed_doi)
    assert no_changes == 0



#!/usr/bin/env python
"""
Import entry from DBLP according to given search input.
"""

import argparse
import logging
import requests
from copy import deepcopy
from pathlib import Path

import bibtex_dblp.database
import bibtex_dblp.dblp_api
import bibtex_dblp.io
from bibtex_dblp.dblp_api import BibFormat, DblpSession


def search_entry(session, search_string, max_search_results, include_arxiv):
    """
    Search an entry for the given search string.
    :param session: DBLP session.
    :param search_string: Search string.
    :param max_search_results: Maximal number of search results to return.
    :param include_arxiv: Whether to include entries from arXiv.
    :return: List of possible entries corresponding to search string, number of total matches
    :raises: HTTPError.
    """
    logging.info("Search: {}".format(search_string))
    search_results = bibtex_dblp.dblp_api.search_publication(session, search_string, max_search_results=max_search_results)
    if include_arxiv:
        return search_results.results, search_results.total_matches
    else:
        valid_results = []
        total_matches = search_results.total_matches
        for res in search_results.results:
            if "CoRR" in str(res.publication):
                total_matches -= 1
            else:
                valid_results.append(res)
        return valid_results, total_matches


def main():
    parser = argparse.ArgumentParser(description="Update entries in bibliography via DBLP.")

    parser.add_argument("infile", help="Input bibtex file", type=Path)
    parser.add_argument("--out", "-o", help="Output bibtex file. If no output file is given, the input file will be overwritten.", type=Path, default=None)
    parser.add_argument("--format", "-f", help="DBLP format type to convert into", type=BibFormat, choices=list(BibFormat), default=BibFormat.condensed)
    parser.add_argument("--max-results", help="Maximal number of search results to display.", type=int, default=30)
    parser.add_argument("--disable-auto", help="Disable automatic selection of publications.", action="store_true")
    parser.add_argument("--include-arxiv", help="Include entries from arXiv in search results.", action="store_true")
    parser.add_argument("--sleep-time", "-t", help="Sleep time (in seconds) between requests. Can prevent errors with too many requests)", type=int, default=5)
    parser.add_argument("--verbose", "-v", help="Print more output", action="store_true")

    args = parser.parse_args()

    logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG if args.verbose else logging.INFO)
    outfile = args.infile if args.out is None else args.out
    bib_format = args.format
    max_search_results = args.max_results
    auto_mode = not args.disable_auto
    include_arxiv = args.include_arxiv

    # Load bibliography
    bib = bibtex_dblp.database.load_from_file(args.infile)
    new_entries = deepcopy(bib.entries)
    # Iterate over all entries
    missing_entries = []
    session = DblpSession(wait_time=args.sleep_time)
    for entry_str, entry in bib.entries.items():
        # Check for id
        dblp_id = bibtex_dblp.dblp_api.extract_dblp_id(entry)
        if dblp_id is not None:
            continue

        if "author" in entry.persons:
            authors = ", ".join([str(author) for author in entry.persons["author"]])
            # Clean-up search string
            authors = authors.replace("{", "")
            authors = authors.replace("}", "")
        else:
            authors = ""
        if "title" in entry.fields:
            title = entry.fields["title"]
            # Clean-up search string
            title = title.replace("{", "")
            title = title.replace("}", "")
        else:
            title = ""

        search_string = "{} {}".format(authors, title)
        try:
            search_results, total_matches = search_entry(session, search_string, max_search_results, include_arxiv)
            if total_matches == 0:
                # Try once again with only the title
                search_results, total_matches = search_entry(session, title, max_search_results, include_arxiv)
        except requests.exceptions.HTTPError as err:
            logging.warning("Search request returned error {}. Skipped this entry.".format(err))
            missing_entries.append(search_string)
            continue

        if total_matches == 0:
            # No luck -> try next entry
            logging.debug("The search returned no matches.")
            missing_entries.append(search_string)
            continue

        if auto_mode and len(search_results) == 1:
            # Select single publication
            publication = search_results[0].publication
            logging.debug("The search returned a single match.")
        else:
            assert len(search_results) > 1
            # Let user select correct publication
            print("The search returned {} matches:".format(total_matches))
            if total_matches > max_search_results:
                print("Displaying only the first {} matches.".format(max_search_results))
            for i in range(len(search_results)):
                result = search_results[i]
                print("({})\t{}".format(i + 1, result.publication))
            # Let user select
            select = bibtex_dblp.io.get_user_number("Select the intended publication (0 to abort): ", 0, total_matches)

            if select == 0:
                missing_entries.append(search_string)
                continue
            publication = search_results[select - 1].publication

        result_dblp = bibtex_dblp.dblp_api.get_bibtex(session, publication.key, bib_format=bib_format)
        data = bibtex_dblp.database.parse_bibtex(result_dblp)
        assert len(data.entries) == 1

        # Update entries
        key = next(iter(data.entries))
        new_entries[entry_str] = data.entries[key]
        logging.debug("Updated entry for '{}'".format(data.entries))

    # Set new entries
    bib.entries = new_entries

    if missing_entries:
        logging.info("The following entries were not found:")
        for m in missing_entries:
            logging.info("- {}".format(m))

    # Write to file
    bibtex_dblp.database.write_to_file(bib, outfile)
    logging.info("Written to {}".format(outfile))


if __name__ == "__main__":
    main()



import re


def search_score(input_string, search_query):
    """
    Return a score indicating how good the given input matches the given search query.
    :param input_string: Input string.
    :param search_query: Search query. Spaces are interpreted as boolean AND. Uses prefix search.
    :return: Score in [0,1] indicating how good the input matches. 0 means no match, 1 means perfect match.
    """
    score = 0
    search_words = search_query.split()
    for word in search_words:
        if re.search(r"\b{}\b".format(word), input_string, re.IGNORECASE) is not None:
            score += 1
    return score / len(search_words)



import bibtex_dblp.dblp_api
from bibtex_dblp.dblp_api import BibFormat


def test_search_publication(dblp_session):
    search_string = "Output-sensitive autocompletion search"
    search_results = bibtex_dblp.dblp_api.search_publication(dblp_session, search_string, max_search_results=30)
    assert search_results.total_matches == 2
    for i in range(len(search_results.results)):
        result = search_results.results[i].publication
        if result.doi == "10.1007/S10791-008-9048-X":
            assert result.title == "Output-sensitive autocompletion search."
            assert result.booktitle is None
            assert result.volume == "11"
            assert result.venue == "Inf. Retr."
            assert result.pages == "269-286"
            assert result.year == 2008
            assert result.type == "Journal Articles"
            assert result.key == "journals/ir/BastMW08"
            assert result.doi == "10.1007/S10791-008-9048-X"
            assert result.ee == "https://doi.org/10.1007/s10791-008-9048-x"
            assert result.url == "https://dblp.org/rec/journals/ir/BastMW08"
            authors = [author.name for author in result.authors]
            assert "Hannah Bast" in authors
            assert "Christian Worm Mortensen" in authors
            assert "Ingmar Weber" in authors
        else:
            assert result.doi == "10.1007/11880561_13"


def test_dblp_bibtex(dblp_session):
    search_string = "Output-sensitive autocompletion search"
    search_results = bibtex_dblp.dblp_api.search_publication(dblp_session, search_string, max_search_results=30)
    assert search_results.total_matches == 2
    result = search_results.results[1].publication
    assert result.doi == "10.1007/11880561_13"

    bibtex_standard = bibtex_dblp.dblp_api.get_bibtex(dblp_session, result.key, bib_format=BibFormat.standard)
    assert "booktitle    = {String Processing and Information Retrieval, 13th International Conference," in bibtex_standard
    assert "{SPIRE} 2006, Glasgow, UK, October 11-13, 2006, Proceedings}" in bibtex_standard

    bibtex_crossref = bibtex_dblp.dblp_api.get_bibtex(dblp_session, result.key, bib_format=BibFormat.crossref)
    assert "crossref     = {DBLP:conf/spire/2006}," in bibtex_crossref
    assert "editor " in bibtex_crossref

    bibtex_condensed = bibtex_dblp.dblp_api.get_bibtex(dblp_session, result.key, bib_format=BibFormat.condensed)
    assert "booktitle    = {{SPIRE}}" in bibtex_condensed
    assert "doi" not in bibtex_condensed

    bibtex_condensed_doi = bibtex_dblp.dblp_api.get_bibtex(dblp_session, result.key, bib_format=BibFormat.condensed_doi)
    assert "booktitle    = {{SPIRE}}" in bibtex_condensed_doi
    assert "doi          = {10.1007/11880561\\_13}" in bibtex_condensed_doi



import logging
import pybtex.database
import re

import bibtex_dblp.dblp_api as dblp_api
import bibtex_dblp.search


def load_from_file(infile):
    """
    Load bibliography from file.
    :param infile: Path of input file.
    :return: Bibliography in pybtex format.
    """
    return pybtex.database.parse_file(infile, bib_format="bibtex")


def write_to_file(bib, outfile):
    """
    Write bibliography to file.
    :param bib: Bibliography in pybtex format.
    :param outfile: Path of output file.
    """
    bib.to_file(outfile, bib_format="bibtex")

    # Perform some custom changes
    content = outfile.read_text(encoding="utf-8")
    # Replace multiple escape characters \\ before by a single one \
    content = re.sub(r"\\{2,}", r"\\", content)
    outfile.write_text(content, encoding="utf-8")


def parse_bibtex(bibtex):
    """
    Parse bibtex string into pybtex format.
    :param bibtex: String containing bibtex information.
    :return: Entry in pybtex format.
    """
    return pybtex.database.parse_string(bibtex, bib_format="bibtex")


def convert_dblp_entries(session, bib, bib_format=dblp_api.BibFormat.condensed):
    """
    Convert bibtex entries according to DBLP bibtex format.
    :param session: DBLP session.
    :param bib: Bibliography in pybtex format.
    :param bib_format: Bibtex format of DBLP.
    :return: Converted bibliography, number of changed entries
    """
    logging.debug("Convert to format '{}'".format(bib_format))
    no_changes = 0
    for entry_str, entry in bib.entries.items():
        # Check for id
        dblp_id = dblp_api.extract_dblp_id(entry)
        if dblp_id is not None:
            logging.debug("Found DBLP id '{}'".format(dblp_id))
            try:
                result_dblp = dblp_api.get_bibtex(session, dblp_id, bib_format=bib_format)
            except dblp_api.InvalidDblpIdException as err:
                logging.warning(str(err) + ". Skipping this entry.")
                continue

            data = parse_bibtex(result_dblp)
            assert len(data.entries) <= 2 if bib_format is dblp_api.BibFormat.crossref else len(data.entries) == 1
            if entry_str not in data.entries:
                # DBLP key is not used as bibtex key -> remember DBLP key
                key = next(iter(data.entries))
                new_entry = data.entries[key]
                new_entry.fields["biburl"] = entry.fields["biburl"]
                bib.entries[entry_str] = new_entry

            else:
                new_entry = data.entries[entry_str]
                # Set new format
                bib.entries[entry_str] = new_entry
                if bib_format is dblp_api.BibFormat.crossref:
                    # Possible second entry
                    for data_key, data_entry in data.entries.items():
                        if data_key != entry_str:
                            if data_key not in bib.entries:
                                bib.entries[data_key] = data_entry
            logging.debug("Set new entry for '{}'".format(entry_str))
            no_changes += 1
    return bib, no_changes


def modify_entries(bib, remove_escapes=False, remove_timestamp=False, remove_biburl=False, remove_bibsource=False):
    """
    Modify bibtex entries.
    :param bib: Bibliography in pybtex format.
    :param remove_escapes: Whether to remove escape characters before underscore in URLs.
    :param remove_timestamp: Whether to remove field 'timestamp'.
    :param remove_biburl: Whether to remove field 'biburl'.
    :param remove_bibsource: Whether to remove field 'bibsource'.
    :return: Modified bibliography.
    """
    if not remove_escapes and not remove_timestamp and not remove_biburl and not remove_bibsource:
        return bib

    for entry_str, entry in bib.entries.items():
        if remove_escapes and "url" in entry.fields:
            entry.fields["url"] = entry.fields["url"].replace("\\_", "_")
        if remove_escapes and "doi" in entry.fields:
            entry.fields["doi"] = entry.fields["doi"].replace("\\_", "_")
        if remove_timestamp and "timestamp" in entry.fields:
            del entry.fields["timestamp"]
        if remove_biburl and "biburl" in entry.fields:
            del entry.fields["biburl"]
        if remove_bibsource and "bibsource" in entry.fields:
            del entry.fields["bibsource"]
    return bib


def search(bib, search_string):
    """
    Search for string in bibliography.
    Only the fields 'author' and 'title' are checked.
    :param bib: Bibliography in pybtex format.
    :param search_string: String to search for.
    :return: List of possible matches of publications with their score.
    """
    results = []
    for _, entry in bib.entries.items():
        if "author" in entry.persons:
            authors = entry.persons["author"]
            author_names = " and ".join([str(author) for author in authors])
        elif "organization" in entry.fields:
            author_names = str(entry.fields["organization"])
        else:
            author_names = ""
        inp = "{}:{}".format(author_names, entry.fields["title"])
        score = bibtex_dblp.search.search_score(inp, search_string)
        if score > 0.5:
            results.append((entry, score))
    results.sort(key=lambda tup: tup[1], reverse=True)
    return results


def print_entry(bib_entry):
    """
    Print pybtex entry.
    :param bib_entry: Pybtex entry.
    :return: String.
    """
    if "author" in bib_entry.persons:
        authors = ", ".join([str(author) for author in bib_entry.persons["author"]])
    elif "organization" in bib_entry.fields:
        authors = str(bib_entry.fields["organization"])
    else:
        authors = ""
    book = ""
    if "booktitle" in bib_entry.fields:
        book = bib_entry.fields["booktitle"]
    if "volume" in bib_entry.fields:
        book += " ({})".format(bib_entry.fields["volume"])
    return "{}:\n\t{} {} {}".format(authors, bib_entry.fields["title"], book, bib_entry.fields["year"])



from conftest import bib_path

from difflib import unified_diff

import bibtex_dblp.database


def test_import_export(tmp_path):
    """
    Check that importing and exporting multiple times yields the same results
    """
    # Import and export
    bib = bibtex_dblp.database.load_from_file(bib_path("ley.bib"))
    tmp_file1 = tmp_path / "export1.bib"
    bibtex_dblp.database.write_to_file(bib, tmp_file1)

    # Compare files
    with open(bib_path("ley.bib")) as f1:
        with open(tmp_file1) as f2:
            diff = list(unified_diff(f1.readlines(), f2.readlines()))
    # The files differ due to different formatting
    # assert not diff, "Files differ:\n{}".format("".join(diff))

    # Import and export
    bib = bibtex_dblp.database.load_from_file(tmp_file1)
    tmp_file2 = tmp_path / "export2.bib"
    bibtex_dblp.database.write_to_file(bib, tmp_file2)

    # Compare files
    with open(tmp_file1) as f1:
        with open(tmp_file2) as f2:
            diff = list(unified_diff(f1.readlines(), f2.readlines()))
    assert not diff, "Files differ:\n{}".format("".join(diff))


def test_modify_entries():
    def count_entries(_bib):
        num_escapes = num_timestamp = num_biburl = num_bibsource = 0
        for entry_str, entry in _bib.entries.items():
            if "url" in entry.fields and "\\_" in entry.fields["url"]:
                num_escapes += 1
            if "doi" in entry.fields and "\\_" in entry.fields["doi"]:
                num_escapes += 1
            if "timestamp" in entry.fields:
                num_timestamp += 1
            if "biburl" in entry.fields:
                num_biburl += 1
            if "bibsource" in entry.fields:
                num_bibsource += 1
        return num_escapes, num_timestamp, num_biburl, num_bibsource

    bib = bibtex_dblp.database.load_from_file(bib_path("ley.bib"))
    assert count_entries(bib) == (2, 9, 9, 9)
    bib = bibtex_dblp.database.modify_entries(bib, remove_escapes=True, remove_timestamp=True, remove_biburl=True, remove_bibsource=True)
    assert count_entries(bib) == (0, 0, 0, 0)
